/*
 * Copyright (c) 2017 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <router.h>
#include <props.h>

#define MAX_NINS 8
#define MAX_NMULT 2
#define MAX_NPROPS 6

typedef struct _plugstate_t plugstate_t;
typedef struct _plughandle_t plughandle_t;

struct _plugstate_t {
	int32_t next_index;
	int32_t prev_index;
	int32_t roll_under_index;
	int32_t roll_over_index;
	int32_t current_index;
	float switch_period;
};

struct _plughandle_t {
	LV2_URID_Map *map;

	const float *audio_in [MAX_NINS][MAX_NMULT];
	float *audio_tmp [MAX_NMULT];
	float *audio_out [MAX_NMULT];
	const LV2_Atom_Sequence *control;
	craft_t notify;

	float target [MAX_NINS];
	float gain [MAX_NINS];
	float ramp_per_sample;

	double sample_rate;

	struct {
		LV2_URID bouncer_nextIndex;
		LV2_URID bouncer_prevIndex;
		LV2_URID bouncer_currentIndex;
	} urid;

	plugstate_t state;
	plugstate_t stash;

	unsigned nmult;

	PROPS_T(props, MAX_NPROPS);
};

static void
_intercept_next_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.next_index)
	{
		handle->state.next_index = false;

		handle->state.current_index += 1;
		if(handle->state.current_index > handle->state.roll_over_index)
			handle->state.current_index = handle->state.roll_under_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.bouncer_nextIndex, &handle->notify.ref);
		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.bouncer_currentIndex, &handle->notify.ref);
	}
}

static void
_intercept_prev_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.prev_index)
	{
		handle->state.prev_index = false;

		handle->state.current_index -= 1;
		if(handle->state.current_index < handle->state.roll_under_index)
			handle->state.current_index = handle->state.roll_over_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.bouncer_prevIndex, &handle->notify.ref);
		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.bouncer_currentIndex, &handle->notify.ref);
	}
}

static void
_intercept_roll_under_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.current_index < handle->state.roll_under_index)
	{
		handle->state.current_index = handle->state.roll_over_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.bouncer_currentIndex, &handle->notify.ref);
	}
}

static void
_intercept_roll_over_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.current_index > handle->state.roll_over_index)
	{
		handle->state.current_index = handle->state.roll_under_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.bouncer_currentIndex, &handle->notify.ref);
	}
}

static void
_intercept_current_index(void *data, int64_t frames,
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	if(handle->state.current_index > handle->state.roll_over_index)
	{
		handle->state.current_index = handle->state.roll_over_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.bouncer_currentIndex, &handle->notify.ref);
	}
	else if(handle->state.current_index < handle->state.roll_under_index)
	{
		handle->state.current_index = handle->state.roll_under_index;

		props_set(&handle->props, &handle->notify.forge, frames,
			handle->urid.bouncer_currentIndex, &handle->notify.ref);
	}
}

static void
_intercept_switch_period(void *data, int64_t frames __attribute__((unused)),
	props_impl_t *impl __attribute__((unused)))
{
	plughandle_t *handle = data;

	handle->ramp_per_sample = (handle->state.switch_period <= 0.f)
		? 1.f
		: (1.f / handle->state.switch_period) / handle->sample_rate;
}

static const props_def_t defs [MAX_NPROPS] = {
	{
		.property = ROUTER__nextIndex,
		.offset = offsetof(plugstate_t, next_index),
		.type = LV2_ATOM__Bool,
		.event_cb = _intercept_next_index
	},
	{
		.property = ROUTER__prevIndex,
		.offset = offsetof(plugstate_t, prev_index),
		.type = LV2_ATOM__Bool,
		.event_cb = _intercept_prev_index
	},
	{
		.property = ROUTER__rollUnderIndex,
		.offset = offsetof(plugstate_t, roll_under_index),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept_roll_under_index
	},
	{
		.property = ROUTER__rollOverIndex,
		.offset = offsetof(plugstate_t, roll_over_index),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept_roll_over_index
	},
	{
		.property = ROUTER__currentIndex,
		.offset = offsetof(plugstate_t, current_index),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept_current_index
	},
	{
		.property = ROUTER__switchPeriod,
		.offset = offsetof(plugstate_t, switch_period),
		.type = LV2_ATOM__Float,
		.event_cb = _intercept_switch_period
	}
};

static LV2_Handle
instantiate(const LV2_Descriptor* descriptor, double rate,
	const char *bundle_path __attribute__((unused)),
	const LV2_Feature *const *features)
{
	plughandle_t *handle = calloc(1, sizeof(plughandle_t));
	if(!handle)
		return NULL;
	mlock(handle, sizeof(plughandle_t));

	if(!strcmp(descriptor->URI, ROUTER__AUDIO_BOUNCER_2))
		handle->nmult = 2;
	else
		handle->nmult = 1;

	LV2_Options_Option *opts = NULL;
	for(unsigned i=0; features[i]; i++)
	{
		if(!strcmp(features[i]->URI, LV2_URID__map))
			handle->map = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_OPTIONS__options))
			opts = features[i]->data;
	}

	if(!handle->map)
	{
		fprintf(stderr,
			"%s: Host does not support urid:map\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	if(!opts)
	{
		fprintf(stderr,
			"%s: Host does not support opts:options\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	_craft_init(&handle->notify, handle->map);

	handle->urid.bouncer_nextIndex = handle->map->map(handle->map->handle,
		ROUTER__nextIndex);
	handle->urid.bouncer_prevIndex = handle->map->map(handle->map->handle,
		ROUTER__prevIndex);
	handle->urid.bouncer_currentIndex = handle->map->map(handle->map->handle,
		ROUTER__currentIndex);

	const LV2_URID bufsz_maxBlockLength = handle->map->map(handle->map->handle,
		LV2_BUF_SIZE__maxBlockLength);

	int32_t max_block_length = 0;
	for(LV2_Options_Option *opt = opts;
		(opt->key != 0) && (opt->value != NULL);
		opt++)
	{
		if(  (opt->key == bufsz_maxBlockLength)
			&& (opt->size == sizeof(int32_t))
			&& (opt->type == handle->notify.forge.Int) )
		{
			max_block_length = *(const int32_t *)opt->value;

			break;
		}
	}

	if(max_block_length == 0)
	{
		fprintf(stderr,
			"%s: Host does not provide bufsz:maxBlockLength\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	if(!props_init(&handle->props, descriptor->URI,
		defs, MAX_NPROPS, &handle->state, &handle->stash,
		handle->map, handle))
	{
		fprintf(stderr, "failed to initialize property structure\n");
		free(handle);
		return NULL;
	}

	for(unsigned k = 0; k < handle->nmult; k++)
	{
		handle->audio_tmp[k] = malloc(max_block_length * sizeof(float)); //FIXME check

		if(!handle->audio_tmp[k])
		{
			while(k-- > 0)
			{
				free(handle->audio_tmp[k]);
			}

			fprintf(stderr, "out-of-memory\n");
			free(handle);
			return NULL;
		}
	}

	handle->sample_rate = rate;

	handle->state.current_index = 1;
	handle->state.roll_over_index = 1;
	handle->state.switch_period = 0.1f;

	return handle;
}

static void
connect_port(LV2_Handle instance, uint32_t port, void *data)
{
	plughandle_t *handle = (plughandle_t *)instance;

	switch(handle->nmult)
	{
		case 1:
		{
			switch(port)
			{
				case 0:
					handle->audio_out[0] = data;
					break;

				case 1:
				case 2:
				case 3:
				case 4:
				case 5:
				case 6:
				case 7:
				case 8:
					handle->audio_in[port-1][0] = (const void *)data;
					break;

				case 9:
					handle->control = (const void *)data;
					break;
				case 10:
					_craft_connect(&handle->notify, data);
					break;
				default:
					break;
			}
		} break;

		case 2:
		{
			switch(port)
			{
				case 0:
					handle->audio_out[0] = data;
					break;
				case 1:
					handle->audio_out[1] = data;
					break;

				case 2:
				case 3:
				case 4:
				case 5:
				case 6:
				case 7:
				case 8:
				case 9:
				case 10:
				case 11:
				case 12:
				case 13:
				case 14:
				case 15:
				case 16:
				case 17:
					handle->audio_in[(port-2)/2][(port-2)%2] = (const void *)data;
					break;

				case 18:
					handle->control = (const void *)data;
					break;
				case 19:
					_craft_connect(&handle->notify, data);
					break;
				default:
					break;
			}

		} break;
	}
}

static void
_run(plughandle_t *handle, uint32_t from, uint32_t to)
{
	// clear audio tmp buffer
	for(unsigned j = from; j < to; j++)
	{
		for(unsigned k = 0; k < handle->nmult; k++)
		{
			handle->audio_tmp[k][j] = 0.f;
		}
	}

	for(int i = 0; i < MAX_NINS; i++)
	{
		// add to tmp
		float gain = handle->gain[i];

		if(gain == handle->target[i])
		{
			if(gain == 1.f)
			{
				for(unsigned j = from; j < to; j++)
				{
					for(unsigned k = 0; k < handle->nmult; k++)
					{
						handle->audio_tmp[k][j] += handle->audio_in[i][k][j];
					}
				}
			}
		}
		else // gain != target
		{
			const float step = handle->target[i] > gain
				? handle->ramp_per_sample
				: -handle->ramp_per_sample;

			for(unsigned j = from; j < to; j++)
			{
				for(unsigned k = 0; k < handle->nmult; k++)
				{
					handle->audio_tmp[k][j] += handle->audio_in[i][k][j] * gain;
				}

				gain += step; // advance gain

				// clip gain
				if(gain < 0.f)
					gain = 0.f;
				else if(gain > 1.f)
					gain = 1.f;
			}

			handle->gain[i] = gain;
		}

		// update target
		handle->target[i] = (i == handle->state.current_index - 1)	
			? 1.f
			: 0.f;
	}

	// copy from tmp to out
	for(unsigned j = from; j < to; j++)
	{
		for(unsigned k = 0; k < handle->nmult; k++)
		{
			handle->audio_out[k][j] = handle->audio_tmp[k][j];
		}
	}
}

static void
run(LV2_Handle instance, uint32_t nsamples)
{
	plughandle_t *handle = instance;

	_craft_in(&handle->notify);

	props_idle(&handle->props, &handle->notify.forge, 0, &handle->notify.ref);

	int64_t from = 0;

	LV2_ATOM_SEQUENCE_FOREACH(handle->control, ev)
	{
		const LV2_Atom *atom = &ev->body;
		const LV2_Atom_Object *obj = (const LV2_Atom_Object *)atom;

		if(props_advance(&handle->props, &handle->notify.forge, ev->time.frames, obj, &handle->notify.ref))
		{
			const int64_t to = ev->time.frames;

			_run(handle, from, to);

			from = to;
		}
	}

	{
		const int64_t to = nsamples;

		_run(handle, from, to);
	}

	_craft_out(&handle->notify);
}

static void
cleanup(LV2_Handle instance)
{
	plughandle_t *handle = instance;

	for(unsigned k = 0; k < handle->nmult; k++)
	{
		free(handle->audio_tmp[k]);
	}

	munlock(handle, sizeof(plughandle_t));
	free(handle);
}

static LV2_State_Status
_state_save(LV2_Handle instance, LV2_State_Store_Function store,
	LV2_State_Handle state, uint32_t flags,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_save(&handle->props, store, state, flags, features);
}

static LV2_State_Status
_state_restore(LV2_Handle instance, LV2_State_Retrieve_Function retrieve,
	LV2_State_Handle state, uint32_t flags,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_restore(&handle->props, retrieve, state, flags, features);
}

static const LV2_State_Interface state_iface = {
	.save = _state_save,
	.restore = _state_restore
};

static const void*
extension_data(const char* uri)
{
	if(!strcmp(uri, LV2_STATE__interface))
		return &state_iface;

	return NULL;
}

const LV2_Descriptor router_audio_bouncer = {
	.URI						= ROUTER__AUDIO_BOUNCER,
	.instantiate		= instantiate,
	.connect_port		= connect_port,
	.activate				= NULL,
	.run						= run,
	.deactivate			= NULL,
	.cleanup				= cleanup,
	.extension_data	= extension_data
};

const LV2_Descriptor router_audio_bouncer_2 = {
	.URI						= ROUTER__AUDIO_BOUNCER_2,
	.instantiate		= instantiate,
	.connect_port		= connect_port,
	.activate				= NULL,
	.run						= run,
	.deactivate			= NULL,
	.cleanup				= cleanup,
	.extension_data	= extension_data
};

const LV2_Descriptor router_cv_bouncer = {
	.URI						= ROUTER__CV_BOUNCER,
	.instantiate		= instantiate,
	.connect_port		= connect_port,
	.activate				= NULL,
	.run						= run,
	.deactivate			= NULL,
	.cleanup				= cleanup,
	.extension_data	= extension_data
};
